#import "VP6Controller.h"
#define LS(str) NSLocalizedString(str,@"")

@implementation VP6Controller

- (id)init
{
	[super init];
	presetArr = [[NSMutableArray alloc] init];
	lock = [[NSLock alloc] init];
	nextPass = NO;
	isEncoding = NO;
	launched = NO;
	attributeDic = [[NSDictionary alloc] initWithObjects:[NSArray arrayWithObject:[NSFont fontWithName:@"Monaco" size:10]] forKeys:[NSArray arrayWithObject:NSFontAttributeName]];
	return self;
}

- (void)dealloc
{
	[presetArr release];
	[lock release];
	[attributeDic release];
	[super dealloc];
}

- (NSMutableDictionary *)currentPresetDictionary
{
	NSMutableDictionary *dic = [[NSMutableDictionary alloc] init];
	[dic setObject:[NSNumber numberWithInt:[o_videoBitrate intValue]] forKey:@"VideoBitrate"];
	[dic setObject:[NSNumber numberWithInt:[[o_pass selectedCell] tag]] forKey:@"Pass"];
	[dic setObject:[NSNumber numberWithInt:[o_minQuantizer intValue]] forKey:@"MinQuantizer"];
	[dic setObject:[NSNumber numberWithInt:[o_maxQuantizer intValue]] forKey:@"MaxQuantizer"];
	[dic setObject:[NSNumber numberWithInt:[o_keyInt intValue]] forKey:@"KeyInt"];
	[dic setObject:[NSNumber numberWithInt:[[o_filter selectedCell] tag]] forKey:@"Filter"];
	[dic setObject:[NSNumber numberWithInt:[o_presetFilter indexOfSelectedItem]] forKey:@"PresetFilter"];
	[dic setObject:[o_customFilter stringValue] forKey:@"CustomFilter"];
	[dic setObject:[NSNumber numberWithInt:[o_variability intValue]] forKey:@"Variability"];
	[dic setObject:[NSNumber numberWithInt:[o_minBitrate intValue]] forKey:@"MinBitrate"];
	[dic setObject:[NSNumber numberWithInt:[o_maxBitrate intValue]] forKey:@"MaxBitrate"];
	[dic setObject:[NSNumber numberWithInt:[o_audioBitrate intValue]] forKey:@"AudioBitrate"];
	[dic setObject:[NSNumber numberWithInt:[[o_mp3Mode selectedCell] tag]] forKey:@"MP3Mode"];
	[dic setObject:[NSNumber numberWithInt:[o_sampleRate indexOfSelectedItem]] forKey:@"SampleRate"];
	return [dic autorelease];
}

- (void)loadPresetFromDictionary:(NSDictionary *)dic
{
	id obj;
	if(obj=[dic objectForKey:@"VideoBitrate"]) {
		[o_videoBitrate setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"Pass"]) {
		[o_pass selectCellWithTag:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"MinQuantizer"]) {
		[o_minQuantizer setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"MaxQuantizer"]) {
		[o_maxQuantizer setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"KeyInt"]) {
		[o_keyInt setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"Filter"]) {
		[o_filter selectCellWithTag:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"PresetFilter"]) {
		[o_presetFilter selectItemAtIndex:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"CustomFilter"]) {
		[o_customFilter setStringValue:obj];
	}
	if(obj=[dic objectForKey:@"Variability"]) {
		[o_variability setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"MinBitrate"]) {
		[o_minBitrate setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"MaxBitrate"]) {
		[o_maxBitrate setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"AudioBitrate"]) {
		[o_audioBitrate setIntValue:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"MP3Mode"]) {
		[o_mp3Mode selectCellWithTag:[obj intValue]];
	}
	if(obj=[dic objectForKey:@"SampleRate"]) {
		[o_sampleRate selectItemAtIndex:[obj intValue]];
	}
	[self filterChanged:o_filter];
	[self passChanged:o_pass];
}

- (void)loadPrefs
{
	NSUserDefaults *pref = [NSUserDefaults standardUserDefaults];
	id obj;
	int i;
	[o_preset removeAllItems];
	if(obj=[pref objectForKey:@"CurrentPreset"]) {
		[self loadPresetFromDictionary:obj];
	}
	if(obj=[pref objectForKey:@"PresetArray"]) {
		[presetArr addObjectsFromArray:obj];
		for(i=0;i<[presetArr count];i++) {
			[o_preset addItemWithTitle:[[presetArr objectAtIndex:i] objectForKey:@"PresetName"]];
		}
		if(obj=[pref objectForKey:@"SelectedPreset"]) {
			[o_preset selectItemAtIndex:[obj intValue]];
		}
	}
	else {
		NSMutableDictionary *dic = [[NSMutableDictionary alloc] init];
		[dic setObject:[NSNumber numberWithInt:500] forKey:@"VideoBitrate"];
		[dic setObject:[NSNumber numberWithInt:1] forKey:@"Pass"];
		[dic setObject:[NSNumber numberWithInt:8] forKey:@"MinQuantizer"];
		[dic setObject:[NSNumber numberWithInt:48] forKey:@"MaxQuantizer"];
		[dic setObject:[NSNumber numberWithInt:300] forKey:@"KeyInt"];
		[dic setObject:[NSNumber numberWithInt:0] forKey:@"Filter"];
		[dic setObject:[NSNumber numberWithInt:0] forKey:@"PresetFilter"];
		[dic setObject:@"" forKey:@"CustomFilter"];
		[dic setObject:[NSNumber numberWithInt:70] forKey:@"Variability"];
		[dic setObject:[NSNumber numberWithInt:40] forKey:@"MinBitrate"];
		[dic setObject:[NSNumber numberWithInt:200] forKey:@"MaxBitrate"];
		[dic setObject:[NSNumber numberWithInt:100] forKey:@"AudioBitrate"];
		[dic setObject:[NSNumber numberWithInt:1] forKey:@"MP3Mode"];
		[dic setObject:[NSNumber numberWithInt:2] forKey:@"SampleRate"];
		[dic setObject:@"512x384 2pass 600kbps" forKey:@"PresetName"];
		[self loadPresetFromDictionary:dic];
		[presetArr addObject:dic];
		[o_preset addItemWithTitle:@"512x384 2pass 600kbps"];
		[dic release];
	}
}

- (void)savePrefs
{
	NSUserDefaults *pref = [NSUserDefaults standardUserDefaults];
	[pref setObject:[self currentPresetDictionary] forKey:@"CurrentPreset"];
	[pref setObject:presetArr forKey:@"PresetArray"];
	[pref setInteger:[o_preset indexOfSelectedItem] forKey:@"SelectedPreset"];
	[pref synchronize];
}

- (NSMutableArray *)arguments
{
	NSMutableArray *arr = [[NSMutableArray alloc] init];
	[arr addObject:@"-of"];
	[arr addObject:@"lavf"];
	[arr addObject:@"-lavfopts"];
	[arr addObject:@"format=flv"];
	[arr addObject:@"-ovc"];
	[arr addObject:@"vfw"];
	if([[o_filter selectedCell] tag] == 0) {
		[arr addObject:@"-sws"];
		[arr addObject:@"9"];
		[arr addObject:@"-vf"];
		if([o_presetFilter indexOfSelectedItem] == 0) [arr addObject:@"flip,scale=512:384"];
		else if([o_presetFilter indexOfSelectedItem] == 1) [arr addObject:@"flip,scale=512:288"];
		else if([o_presetFilter indexOfSelectedItem] == 2) [arr addObject:@"flip,scale=512:288:::3,expand=0:-96:0:48"];
	}
	else {
		if(![[o_customFilter stringValue] isEqualToString:@""]) [arr addObjectsFromArray:[[o_customFilter stringValue] componentsSeparatedByString:@" "]];
	}
	[arr addObject:@"-oac"];
	[arr addObject:@"mp3lame"];
	[arr addObject:@"-lameopts"];
	if([[o_mp3Mode selectedCell] tag] == 0) [arr addObject:[NSString stringWithFormat:@"cbr:preset=%d",[o_audioBitrate intValue]]];
	else [arr addObject:[NSString stringWithFormat:@"preset=%d",[o_audioBitrate intValue]]];
	[arr addObject:@"-af"];
	if([o_sampleRate indexOfSelectedItem] == 0) {
		[arr addObject:@"lavcresample=11025"];
		[arr addObject:@"-srate"];
		[arr addObject:@"11025"];
	}
	else if([o_sampleRate indexOfSelectedItem] == 1) {
		[arr addObject:@"lavcresample=22050"];
		[arr addObject:@"-srate"];
		[arr addObject:@"22050"];
	}
	else if([o_sampleRate indexOfSelectedItem] == 2) {
		[arr addObject:@"lavcresample=44100"];
		[arr addObject:@"-srate"];
		[arr addObject:@"44100"];
	}
	[arr addObject:@"-o"];
	[arr addObject:[o_outputFile stringValue]];
	[arr addObject:[o_inputFile stringValue]];
	//NSLog([arr description]);
	return [arr autorelease];
}

- (NSMutableArray *)argumentsNoAudio
{
	NSMutableArray *arr = [[NSMutableArray alloc] init];
	[arr addObject:@"-of"];
	[arr addObject:@"lavf"];
	[arr addObject:@"-lavfopts"];
	[arr addObject:@"format=flv"];
	[arr addObject:@"-ovc"];
	[arr addObject:@"vfw"];
	if([[o_filter selectedCell] tag] == 0) {
		[arr addObject:@"-sws"];
		[arr addObject:@"9"];
		[arr addObject:@"-vf"];
		if([o_presetFilter indexOfSelectedItem] == 0) [arr addObject:@"flip,scale=512:384"];
		else if([o_presetFilter indexOfSelectedItem] == 1) [arr addObject:@"flip,scale=512:288"];
		else if([o_presetFilter indexOfSelectedItem] == 2) [arr addObject:@"flip,scale=512:288:::3,expand=0:-96:0:48"];
	}
	else {
		if(![[o_customFilter stringValue] isEqualToString:@""]) [arr addObjectsFromArray:[[o_customFilter stringValue] componentsSeparatedByString:@" "]];
	}
	[arr addObject:@"-nosound"];
	[arr addObject:@"-o"];
	[arr addObject:[o_outputFile stringValue]];
	[arr addObject:[o_inputFile stringValue]];
	//NSLog([arr description]);
	return [arr autorelease];
}

- (void)launchEncoderWithMCF:(NSString *)mcfFile
{
	int bitrate,variability,maxrate,minrate,keyint,maxquant,minquant;
	task = [[NSTask alloc] init];
	id pipe1 = [NSPipe pipe];
	id pipe2 = [NSPipe pipe];
	[task setStandardOutput: pipe1];
	[task setStandardError: pipe2];
	[task setLaunchPath:[[NSBundle mainBundle] pathForResource:@"mencoder" ofType:nil inDirectory:nil]];
	[task setCurrentDirectoryPath:[[NSBundle mainBundle] resourcePath]];
	NSMutableArray *args = nextPass ? [self argumentsNoAudio] : [self arguments];
	[args addObject:@"-xvfwopts"];
	[args addObject:[NSString stringWithFormat:@"codec=vp6vfw.dll:compdata=%@",mcfFile]];
	[task setArguments:args];
	[[NSNotificationCenter defaultCenter] addObserver:self 
											 selector:@selector(taskDidEnd:) 
												 name: @"NSTaskDidTerminateNotification" 
											   object: nil];
	FILE *fp = fopen([[[NSBundle mainBundle] pathForResource:mcfFile ofType:nil inDirectory:nil] UTF8String],"r+b");
	
	bitrate = [o_videoBitrate intValue];
	variability = [o_variability intValue];
	maxrate = [o_maxBitrate intValue];
	minrate = [o_minBitrate intValue];
	keyint = [o_keyInt intValue];
	maxquant = [o_maxQuantizer intValue];
	minquant = [o_minQuantizer intValue];
	
	fseek(fp,0x14,SEEK_SET);
	fwrite(&bitrate,4,1,fp);
	fseek(fp,0x18,SEEK_SET);
	fwrite(&maxquant,4,1,fp);
	fseek(fp,0x38,SEEK_SET);
	fwrite(&keyint,4,1,fp);
	fseek(fp,0x94,SEEK_SET);
	fwrite(&minquant,4,1,fp);
	fseek(fp,0xa8,SEEK_SET);
	fwrite(&variability,4,1,fp);
	fseek(fp,0xac,SEEK_SET);
	fwrite(&maxrate,4,1,fp);
	fseek(fp,0xb0,SEEK_SET);
	fwrite(&minrate,4,1,fp);
	
	fclose(fp);
	
	[o_progress setDoubleValue:0.0];
	isEncoding = YES;
	[NSThread detachNewThreadSelector:@selector(watchOutput) toTarget:self withObject:nil];
	[NSThread detachNewThreadSelector:@selector(watchError) toTarget:self withObject:nil];
	[task launch];
}

- (void)setInputFile:(NSString *)path
{
	[o_inputFile setStringValue:path];
	[o_outputFile setStringValue:[[path stringByDeletingPathExtension] stringByAppendingPathExtension:@"flv"]];
}

- (BOOL)isEncoding
{
	return isEncoding;
}

- (IBAction)filterChanged:(id)sender
{
	if([[sender selectedCell] tag] == 0) {
		[o_presetFilter setEnabled:YES];
		[o_customFilter setEnabled:NO];
	}
	else {
		[o_presetFilter setEnabled:NO];
		[o_customFilter setEnabled:YES];
	}
}

- (IBAction)mp3ModeChanged:(id)sender
{
}

- (IBAction)passChanged:(id)sender
{
	if([[sender selectedCell] tag] == 0) {
		[o_text1 setTextColor:[NSColor grayColor]];
		[o_text2 setTextColor:[NSColor grayColor]];
		[o_text3 setTextColor:[NSColor grayColor]];
		[o_text4 setTextColor:[NSColor grayColor]];
		[o_variability setEnabled:NO];
		[o_minBitrate setEnabled:NO];
		[o_maxBitrate setEnabled:NO];
	}
	else {
		[o_text1 setTextColor:[NSColor blackColor]];
		[o_text2 setTextColor:[NSColor blackColor]];
		[o_text3 setTextColor:[NSColor blackColor]];
		[o_text4 setTextColor:[NSColor blackColor]];
		[o_variability setEnabled:YES];
		[o_minBitrate setEnabled:YES];
		[o_maxBitrate setEnabled:YES];
	}
}

- (IBAction)presetChanged:(id)sender
{
	[self loadPresetFromDictionary:[presetArr objectAtIndex:[o_preset indexOfSelectedItem]]];
}

- (IBAction)addPreset:(id)sender
{
	[NSApp beginSheet:o_presetDialog
	   modalForWindow:o_mainWindow
		modalDelegate:self
	   didEndSelector:@selector(sheetDidEnd:returnCode:contextInfo:)
		  contextInfo:@"AddPreset"];
}

- (IBAction)removePreset:(id)sender
{
	int idx = [o_preset indexOfSelectedItem];
	[presetArr removeObjectAtIndex:idx];
	[o_preset removeItemAtIndex:idx];
	if([presetArr count] == 1) [o_removePreset setEnabled:NO];
}

- (IBAction)updatePreset:(id)sender
{
	[[presetArr objectAtIndex:[o_preset indexOfSelectedItem]] addEntriesFromDictionary:[self currentPresetDictionary]];
}

- (IBAction)endEdit:(id)sender
{
	if([sender tag] == 1) {
		if([[o_presetName stringValue] isEqualToString:@""]) {
			NSRunCriticalAlertPanel(LS(@"error"), LS(@"name is empty"), LS(@"Cancel"), nil, nil);
			return;
		}
		int i;
		NSArray *arr = [o_preset itemTitles];
		for(i=0;i<[arr count];i++) {
			if([[o_presetName stringValue] isEqualToString:[arr objectAtIndex:i]]) {
				NSRunCriticalAlertPanel(LS(@"error"), LS(@"duplicate name"), LS(@"Cancel"), nil, nil);
				return;
			}
		}
	}
	[NSApp endSheet:o_presetDialog returnCode:[sender tag]];
	[o_presetDialog close];
}

- (IBAction)beginEncode:(id)sender
{
	if(!isEncoding) {
		[o_audioBitrate validateEditing];
		[o_customFilter validateEditing];
		[o_keyInt validateEditing];
		[o_maxBitrate validateEditing];
		[o_maxQuantizer validateEditing];
		[o_minBitrate validateEditing];
		[o_minQuantizer validateEditing];
		[o_variability validateEditing];
		[o_videoBitrate validateEditing];
		
		[o_audioBitrate setEnabled:NO];
		[o_customFilter setEnabled:NO];
		[o_filter setEnabled:NO];
		[o_keyInt setEnabled:NO];
		[o_maxBitrate setEnabled:NO];
		[o_maxQuantizer setEnabled:NO];
		[o_minBitrate setEnabled:NO];
		[o_minQuantizer setEnabled:NO];
		[o_mp3Mode setEnabled:NO];
		[o_pass setEnabled:NO];
		[o_preset setEnabled:NO];
		[o_presetFilter setEnabled:NO];
		[o_sampleRate setEnabled:NO];
		[o_variability setEnabled:NO];
		[o_videoBitrate setEnabled:NO];
		[o_progress setHidden:NO];
		[o_startButton setTitle:LS(@"stop")];
		[[[o_log textStorage] mutableString] setString:@""];
		[o_logWindow orderFront:nil];
		if([[o_pass selectedCell] tag] == 0) {
			nextPass = NO;
			[o_status setStringValue:LS(@"encoding")];
			[self launchEncoderWithMCF:@"1passVBR.mcf"];
		}
		else {
			nextPass = YES;
			[o_status setStringValue:LS(@"analyzing")];
			[self launchEncoderWithMCF:@"2passVBR_1st.mcf"];
		}
	}
	else {
		[task terminate];
	}
}

- (IBAction)chooseInputFile:(id)sender
{
	if(isEncoding) return;
	NSOpenPanel *op = [NSOpenPanel openPanel];
	[op setCanChooseDirectories:NO];
	[op setCanChooseFiles:YES];
	[op setAllowsMultipleSelection:NO];
	
	int ret = [op runModalForDirectory:nil file:nil types:nil];
	if(ret != NSOKButton) return;
	
	[o_inputFile setStringValue:[op filename]];
	[o_outputFile setStringValue:[[[op filename] stringByDeletingPathExtension] stringByAppendingPathExtension:@"flv"]];
}

- (IBAction)chooseOutputFile:(id)sender
{
	if(isEncoding) return;
	NSSavePanel *sv = [NSSavePanel savePanel];
	
	int ret = [sv runModalForDirectory:nil file:[[o_outputFile stringValue] length] ? [[o_outputFile stringValue] lastPathComponent] : nil];
	if(ret != NSOKButton) return;
	
	[o_outputFile setStringValue:[sv filename]];
}

- (IBAction)showLog:(id)sender
{
	[o_logWindow makeKeyAndOrderFront:nil];
}

- (void)applicationDidFinishLaunching: (NSNotification *)notification
{
	NSFileManager *fm = [NSFileManager defaultManager];
	if(![fm fileExistsAtPath:@"/usr/local/lib/codecs/vp6vfw.dll"]) {
		[o_urlStr setAllowsEditingTextAttributes: YES];
		[o_urlStr setAttributedStringValue:[NSMutableAttributedString hyperlinkFromString:@"http://www.mplayerhq.hu/design7/dload.html" withURL:[NSURL URLWithString:@"http://www.mplayerhq.hu/design7/dload.html"] alignment:NSCenterTextAlignment]];
		[o_errorPanel center];
		[o_errorPanel makeKeyAndOrderFront:nil];
		return;
	}
	[self loadPrefs];
	[self filterChanged:o_filter];
	[self passChanged:o_pass];
	[o_progress setHidden:YES];
	if([presetArr count] == 1) [o_removePreset setEnabled:NO];
	[o_status setStringValue:LS(@"idle")];
	[o_bps setStringValue:@""];
	NSArray* array = [NSArray arrayWithObject:NSFilenamesPboardType];
	[o_mainWindow registerForDraggedTypes:array];
	[o_mainWindow makeKeyAndOrderFront:nil];
	launched = YES;
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	if(launched) [self savePrefs];
	return YES;
}

- (void)sheetDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	if(!returnCode) return;
	NSMutableDictionary *dic = [self currentPresetDictionary];
	[dic setObject:[o_presetName stringValue] forKey:@"PresetName"];
	[presetArr addObject:dic];
	[o_preset addItemWithTitle:[o_presetName stringValue]];
	[o_preset selectItemAtIndex:[presetArr count]-1];
	[o_removePreset setEnabled:YES];
}
/*
- (void)getData: (NSNotification *)aNotification
{
	NSData *data = [[aNotification userInfo] objectForKey:NSFileHandleNotificationDataItem];
	if ([data length]) {
		NSString *str = [[[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding] autorelease];
		NSLog(str);
	}
	if([task isRunning]) [[aNotification object] readInBackgroundAndNotify];
}
*/
- (void)watchOutput
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	NSFileHandle *handle = [[[task standardOutput] fileHandleForReading] retain];
	unsigned char *buffer = (unsigned char*)malloc(1024);
	int p=0;
	unsigned sec = time(NULL);
	
	while(isEncoding) {
		NSAutoreleasePool *pool2 = [[NSAutoreleasePool alloc] init];
		
		NSData *data = [handle availableData];
		if(data && [data length] > 0) {
			int i;
			unsigned char *ptr = (unsigned char *)[data bytes];
			for(i=0;i<[data length];i++) {
				buffer[p++] = ptr[i];
				if(ptr[i] == '\n' || p == 1024 || ptr[i] == '\r') {
					buffer[p-1] = 0;
					if(!strncmp("Pos:",(char *)buffer,4)) {
						int j=0;
						while((char)buffer[j] != '(') j++;
						[o_progress setDoubleValue:atoi((char *)buffer+j+1)];
						if(!nextPass && time(NULL)-sec>=1) {
							while((char)buffer[j] != '[') j++;
							int video = atoi((char *)buffer+j+1);
							while((char)buffer[j] != ':') j++;
							int audio = atoi((char *)buffer+j+1);
							[o_bps setStringValue:[NSString stringWithFormat:@"(%@ %d+%d kbps)",LS(@"current"),video,audio]];
							sec = time(NULL);
						}
					}
					else {
						if(strlen((char *)buffer)) {
							NSAttributedString *str = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%@\n",[NSString stringWithUTF8String:(char *)buffer]] attributes: attributeDic];
							[lock lock];
							//int start = [[[o_log textStorage] mutableString] length];
							[[o_log textStorage]  appendAttributedString:str];
							//[[o_log textStorage] setAttributes:attributeDic range:NSMakeRange(start,[str length]+1)];
							[o_log autoscroll:nil];
							[lock unlock];
							[str release];
						}
					}
					p = 0;
				}
			}
		}
		
		[pool2 release];
	}
	//NSLog(@"thread end");
	free(buffer);
	[handle release];
	[pool release];
}

- (void)watchError
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	NSFileHandle *handle = [[[task standardError] fileHandleForReading] retain];
	unsigned char *buffer = (unsigned char*)malloc(1024);
	int p=0;
	
	while(isEncoding) {
		NSAutoreleasePool *pool2 = [[NSAutoreleasePool alloc] init];
		
		NSData *data = [handle availableData];
		if(data && [data length] > 0) {
			int i;
			unsigned char *ptr = (unsigned char *)[data bytes];
			for(i=0;i<[data length];i++) {
				buffer[p++] = ptr[i];
				if(ptr[i] == '\n' || p == 1024 || ptr[i] == '\r') {
					buffer[p-1] = 0;
					if(strlen((char *)buffer)) {
						NSAttributedString *str = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%@\n",[NSString stringWithUTF8String:(char *)buffer]] attributes: attributeDic];
						[lock lock];
						//int start = [[[o_log textStorage] mutableString] length];
						[[o_log textStorage] appendAttributedString:str];
						//[[o_log textStorage] setAttributes:attributeDic range:NSMakeRange(start,[str length]+1)];
						[o_log autoscroll:nil];
						[lock unlock];
						[str release];
					}
					p = 0;
				}
			}
		}
		
		[pool2 release];
	}
	//NSLog(@"thread end");
	free(buffer);
	[handle release];
	[pool release];
}

- (void)taskDidEnd: (NSNotification *)aNotification
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	isEncoding = NO;
	int status = [task terminationStatus];
	[task release];
	if(nextPass && status == 0) {
		nextPass = NO;
		[o_status setStringValue:LS(@"encoding")];
		[self launchEncoderWithMCF:@"2passVBR_2nd.mcf"];
	}
	else {
		[o_progress setHidden:YES];
		[o_startButton setTitle:LS(@"begin")];
		[o_audioBitrate setEnabled:YES];
		[o_filter setEnabled:YES];
		[o_keyInt setEnabled:YES];
		[o_maxQuantizer setEnabled:YES];
		[o_minQuantizer setEnabled:YES];
		[o_mp3Mode setEnabled:YES];
		[o_pass setEnabled:YES];
		[o_preset setEnabled:YES];
		[o_sampleRate setEnabled:YES];
		[o_videoBitrate setEnabled:YES];
		[self filterChanged:o_filter];
		[self passChanged:o_pass];
		[o_status setStringValue:LS(@"idle")];
		[o_bps setStringValue:@""];
		[NSApp requestUserAttention:NSInformationalRequest];
		NSBeep();
	}
}

@end


@implementation NSAttributedString (Hyperlink)
+ (id)hyperlinkFromString:(NSString*)inString withURL:(NSURL*)aURL alignment:(NSTextAlignment)alignment
{
    NSMutableAttributedString* attrString = [[NSMutableAttributedString alloc] initWithString: inString];
    NSRange range = NSMakeRange(0, [attrString length]);
	
    [attrString beginEditing];
    [attrString addAttribute:NSLinkAttributeName value:[aURL absoluteString] range:range];
	
    [attrString addAttribute:NSForegroundColorAttributeName value:[NSColor blueColor] range:range];
    [attrString addAttribute:NSUnderlineStyleAttributeName value:[NSNumber numberWithInt:NSSingleUnderlineStyle] range:range];
	[attrString addAttribute:NSFontAttributeName value:[NSFont systemFontOfSize:11] range:range];
	NSMutableParagraphStyle *style = [[NSParagraphStyle defaultParagraphStyle] mutableCopy];
	[style setAlignment:alignment];
	[attrString addAttribute:NSParagraphStyleAttributeName value:style range:range];
	[style release];
	
    [attrString endEditing];
	
    return [attrString autorelease];
}
@end